module Main exposing (..)

import Browser
import Html exposing (..)
import Html.Attributes exposing (..)
import Html.Events exposing (..)
import Svg exposing (svg)
import Svg.Attributes
import Character exposing (Character)
import Stats exposing (Stat)
import BoundedInt exposing (BoundedInt)
import Stats exposing (AttributeCategory)
import Creation exposing (CreationPoints)
import OrderedAttributeCategories as OAC exposing (OAC)

main : Program () Model Msg
main = 
    Browser.document
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        }

init : () -> ( Model, Cmd Msg )
init _ =
    (   { character = Character.new
        , modal = Nothing
        , modalValue = ""
        }
    , Cmd.none
    )

type alias Model = 
    { character : Character
    , modal : Maybe ModalType
    , modalValue : String
    }

type Msg 
    = OpenModal ModalType
    | CloseModal
    | ChangeModalValue String
    | SaveModalChange ModalType
    | ChangeStat Stat
    | ChangePriority AttributeCategory Int

type ModalType
    = Name

update : Msg -> Model -> ( Model, Cmd Msg )
update msg model = 
    case msg of
        OpenModal modalType -> 
            (   { model 
                | modal = ( Just modalType ) 
                , modalValue = modalValue modalType model.character
                }
            , Cmd.none
            )
        CloseModal ->
            ( { model | modal = Nothing }
            , Cmd.none
            )
        ChangeModalValue value ->
            ( { model | modalValue = value }
            , Cmd.none
            )
        SaveModalChange modalType ->
            case modalType of
                Name ->
                    (   { model 
                        | character = Character.changeName model.character model.modalValue 
                        , modal = Nothing
                        }
                    , Cmd.none
                    )
        ChangeStat stat ->
            (   { model 
                | character = Character.changeStat model.character stat 
                }
            , Cmd.none
            )
        ChangePriority category value ->
            (   { model 
                | character = Character.changePriority model.character category value
                }
            , Cmd.none
            )

view : Model -> Browser.Document Msg
view model =
    { title = "Character Sheet"
    , body =
        [ div [ class "container is-max-desktop" ] 
            [ viewEditableText Name model.character.name
            , viewAttributes model.character.creationPoints model.character.creationPoints.attributes model.character.stats
            , viewSpheres model.character.creationPoints ( Stats.spheresInList model.character.stats )
            , viewStat  ( Stats.areteInList model.character.stats )
            , viewModal model
            , div [] [ text ( "freebie: " ++ String.fromInt model.character.creationPoints.freebie ) ]
            , div [] [ text ( "spheres: " ++ String.fromInt model.character.creationPoints.spheres ) ]
            ]
        ]
    }

viewEditableText : ModalType -> String -> Html Msg
viewEditableText modalType content =
    div []
        [ span [ class "icon-text" ]
            [ span [] [ text content ]
            , span 
                [ class "icon" 
                ] 
                [ i 
                    [ class "fas fa-pen"
                    , onClick ( OpenModal modalType ) 
                    ] []
                ]
            ]
        ]

viewAttributes : CreationPoints -> OAC -> List Stat -> Html Msg
viewAttributes cp oac attributes =
    div [ class "has-text-centered" ]
        [ p [] [ text "Attributes" ]
        , div [ class "columns is-centered" ]
            [ viewAttributeCol oac "Physical" Stats.Physical attributes
            , viewAttributeCol oac "Social" Stats.Social attributes
            , viewAttributeCol oac "Mental" Stats.Mental attributes
            ]
        ]

viewAttributeCol : OAC -> String -> AttributeCategory -> List Stat -> Html Msg
viewAttributeCol oac name category attributes =
    let 
        priority = OAC.indexOf oac.order category
        points = OAC.categoryValue oac category
    in
        div [ class "column is-one-fifth" ]
            (   [ p [] [ text name ]
                , creationInfo points
                , div [ class "field has-addons is-fullwidth" ]
                    [ p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if priority == 0 then " is-primary" else "" )
                            , onClick ( ChangePriority category 0 )
                            ] [ text "I" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if priority == 1 then " is-primary" else "" )
                            , onClick ( ChangePriority category 1 )
                            ] [ text "II" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if priority == 2 then " is-primary" else "" )
                            , onClick ( ChangePriority category 2 )
                            ] [ text "III" ]
                        ]
                    ]
                ]
                ++ 
                ( List.map viewStat ( Stats.attributesOfCategoryInList category attributes )
                )
            )

viewSpheres : CreationPoints -> List Stat -> Html Msg
viewSpheres cp spheres =
    div [ class "has-text-centered" ]
        [ p [] [ text "Spheres" ]
        , div [ class "columns is-centered" ]
            [ div [ class "column is-three-fifths" ]
                [ creationInfo cp.spheres
                ]
            ]
        , div [ class "columns is-centered" ]
            [ div [ class "column is-one-fifth" ] 
                ( List.map viewStat ( Stats.everyThirdStat spheres 0 )
                )
            , div [ class "column is-one-fifth" ] 
                ( List.map viewStat ( Stats.everyThirdStat spheres 1 )
                )
            , div [ class "column is-one-fifth" ] 
                ( List.map viewStat ( Stats.everyThirdStat spheres 2 )
                )
            ]
        ]

creationInfo : Int -> Html Msg
creationInfo points = 
    let
        notificationColor : String
        notificationColor =
            case compare points 0 of
                GT -> "is-warning"
                EQ -> "is-success"
                LT -> "is-danger"
    in
        div [ class ( "notification " ++ notificationColor ) ]
            [ text ( ( String.fromInt points ) ++ " remaining" )
            ]

viewStat : Stat -> Html Msg
viewStat stat =
    div [ class "notification" ]
        [ p [] [ text stat.name ]
        , p [] [ viewDots stat ]
        ]
        

viewDots : Stat -> Html Msg
viewDots stat =
    let
        dotSize = 12
        dotGap = 4
        dotOutline = 0.05 * dotSize
        marginalIncrease = ( dotSize + dotGap ) / 2
        svgOutlineMargin = ( sqrt ( 2 * ( dotOutline ^ 2 ) ) ) / 2

        rangeList = List.range 0 ( stat.dots.max - 1 )

        isFilledList = rangeList |> List.map ( (>) stat.dots.value )
    in
        svg 
            [ Svg.Attributes.width  ( String.fromFloat ( dotSize + marginalIncrease * toFloat ( stat.dots.max - 1 ) + 2 * svgOutlineMargin ) )
            , Svg.Attributes.height ( String.fromFloat ( dotSize + marginalIncrease + 2 * svgOutlineMargin ) )
            ]
            ( rangeList
                |> List.map ( toDotPosition dotSize dotGap svgOutlineMargin )
                |> List.map3 ( viewDot stat dotSize dotOutline ) isFilledList rangeList
            )

toDotPosition : Float -> Float -> Float -> Int -> ( Float, Float )
toDotPosition size gap offset id =
    let
        marginalIncrease = ( size + gap ) / 2
    in
        ( toFloat id * marginalIncrease + offset
        , toFloat ( remainderBy 2 ( id + 1 ) ) * marginalIncrease + offset
        )

viewDot : Stat -> Float -> Float -> Bool -> Int -> ( Float, Float ) -> Html Msg
viewDot stat size strokeWidth isFilled id position =
    Svg.polygon 
        [ Svg.Attributes.points 
            (   [ ( 0.5, 0.0 )
                , ( 1.0, 0.5 )
                , ( 0.5, 1.0 )
                , ( 0.0, 0.5 )
                ]
                |> List.map ( Tuple.mapBoth   ( (*) size ) ( (*) size ) )
                |> List.map ( Tuple.mapFirst  ( (+) ( Tuple.first  position ) ) )
                |> List.map ( Tuple.mapSecond ( (+) ( Tuple.second position ) ) )
                |> List.map tupleToString
                |> List.intersperse " "
                |> String.concat
            )
        , Svg.Attributes.fill "var(--bulma-body-color)"
        , Svg.Attributes.fillOpacity ( boolAsIntString isFilled )
        , Svg.Attributes.stroke "var(--bulma-body-color)"
        , Svg.Attributes.strokeWidth ( String.fromFloat strokeWidth )
        , onClick ( ChangeStat { stat | dots = BoundedInt.changeTo stat.dots ( valueOrReset stat.dots ( id + 1 ) ) } )
        ] []

tupleToString : ( Float, Float ) -> String
tupleToString tuple =
    let
        first  = String.fromFloat ( Tuple.first  tuple )
        second = String.fromFloat ( Tuple.second tuple )
    in
        first ++ "," ++ second

boolAsIntString : Bool -> String
boolAsIntString bool =
    if bool
    then "1"
    else "0"

valueOrReset : BoundedInt -> Int -> Int
valueOrReset dots value =
    if dots.value == value
    then dots.min
    else value

viewModal : Model -> Html Msg
viewModal model =
    case model.modal of
        Nothing -> div [ class "modal" ] []
        Just modalType ->
            div [ class "modal is-active" ]
                [ div 
                    [ class "modal-background" 
                    , onClick CloseModal
                    ] []
                , div [ class "modal-card" ] 
                    [ header [ class "modal-card-head" ]
                        [ p [ class "modal-card-title" ] [ text ( "Change " ++ ( modalTitle modalType ) ) ]
                        ]
                    , section [ class "modal-card-body" ] 
                        [ p [] 
                            [ input 
                                [ class "input" 
                                , type_ "text"
                                , placeholder ( "Enter New " ++ ( modalTitle modalType ) )
                                , value model.modalValue
                                , onInput ( \x -> ChangeModalValue x )
                                ] []
                            ]
                        ]
                    , footer [ class "modal-card-foot" ] 
                        [ div [ class "buttons" ] 
                            [ button 
                                [ class "button is-success"
                                , onClick ( SaveModalChange modalType )
                                ] [ text "Save Changes" ]
                            , button 
                                [ class "button"
                                , onClick CloseModal
                                ] [ text "Cancel" ]
                            ]
                        ]
                    ]
                ]

modalTitle : ModalType -> String
modalTitle modalType =
    case modalType of
        Name -> "Character Name"

modalValue : ModalType -> Character -> String
modalValue modalType character =
    case modalType of
        Name -> character.name

subscriptions : Model -> Sub Msg
subscriptions _ =
  Sub.none

-- aria functions

ariaLabel : String -> Attribute msg
ariaLabel value =
    Html.Attributes.attribute "aria-label" value

ariaHaspopup : String -> Attribute msg
ariaHaspopup value =
    Html.Attributes.attribute "aria-haspopup" value

ariaControls : String -> Attribute msg
ariaControls value =
    Html.Attributes.attribute "aria-controls" value

ariaHidden : String -> Attribute msg
ariaHidden value =
    Html.Attributes.attribute "aria-hidden" value

ariaRole : String -> Attribute msg
ariaRole value =
    Html.Attributes.attribute "role" value